#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <limits.h>
#include <elf.h>
#include "../../kexec.h"
#include "../../kexec-elf.h"
#include "../../kexec-syscall.h"
#include "kexec-sparc64.h"
#include "crashdump-sparc64.h"
#include "../kexec/crashdump.h"

static int memory_ranges;
static struct memory_range *crash_memory_range;
struct memory_range crash_reserved_mem;

int get_crash_memory_ranges(struct memory_range **range, int *ranges)
{
	const char *iomem = proc_iomem();
        char line[MAX_LINE];
        FILE *fp;
        unsigned long start, end;

	crash_memory_range = xmalloc(sizeof(struct memory_range) *
				max_memory_ranges);
        fp = fopen(iomem, "r");
        if (!fp) {
                fprintf(stderr, "Cannot open %s: %s\n",
                        iomem, strerror(errno));
                return -1;
        }
	while(fgets(line, sizeof(line), fp) != 0) {
		char *str;
		int type, consumed, count;
		if (memory_ranges >= max_memory_ranges)
			break;
		count = sscanf(line, "%lx-%lx : %n",
				&start, &end, &consumed);
		str = line + consumed;
		if (count != 2)
			continue;

		if (memcmp(str, "System RAM\n", 11) == 0) {
			type = RANGE_RAM;
		} else if (memcmp(str, "Crash kernel\n", 13) == 0) {
			/* Reserved memory region. New kernel can
			 * use this region to boot into. */
			crash_reserved_mem.start = start;
			crash_reserved_mem.end = end;
			crash_reserved_mem.type = RANGE_RAM;
			continue;
		} else if (memcmp(str, "Uncached RAM\n", 13) == 0) {
			type = RANGE_UNCACHED;
		} else {
			continue;
		}
		crash_memory_range[memory_ranges].start = start;
		crash_memory_range[memory_ranges].end = end;
		crash_memory_range[memory_ranges].type = type;
		memory_ranges++;
	}
        fclose(fp);
	*range = crash_memory_range;
	*ranges = memory_ranges;
	return 0;
}

int is_crashkernel_mem_reserved(void)
{
	uint64_t start, end;

	return parse_iomem_single("Crash kernel\n", &start,
				  &end) == 0 ?  (start != end) : 0;
}

/* We change the PT_LOAD p_addr to be the first crash segment region.
 * Bah, should there ever be more than one.
 */
void kexec_crash_move_pt_load(struct mem_ehdr *ehdr)
{
	int i;
	unsigned long start, end;

	start = crash_reserved_mem.start;
	end = crash_reserved_mem.end;

	for (i = 0; i < ehdr->e_phnum; i++) {
		struct mem_phdr *phdr = &ehdr->e_phdr[i];

		if (phdr->p_type != PT_LOAD)
			continue;
		else {
			phdr->p_paddr = start + (phdr->p_vaddr - KERNBASE);
			if ((phdr->p_paddr + phdr->p_memsz) > end)
				die("Crash Area too small");
		}
	}
}

/* Retrieve kernel _stext symbol virtual address from /proc/kallsyms */
static unsigned long long get_kernel_stext_sym(void)
{
	const char *kallsyms = "/proc/kallsyms";
	const char *stext = "_stext";
	char sym[128];
	char line[128];
	FILE *fp;
	unsigned long long vaddr;
	char type;

	fp = fopen(kallsyms, "r");
	if (!fp) {
		fprintf(stderr, "Cannot open %s\n", kallsyms);
		return 0;
	}

	while(fgets(line, sizeof(line), fp) != NULL) {
		if (sscanf(line, "%Lx %c %s", &vaddr, &type, sym) != 3)
			continue;
		if (strcmp(sym, stext) == 0) {
			dbgprintf("kernel symbol %s vaddr = %16llx\n", stext, vaddr);
			return vaddr;
		}
	}

	fprintf(stderr, "Cannot get kernel %s symbol address\n", stext);
	return 0;
}

static int get_kernel_region(struct crash_elf_info *elf_info)
{
	uint64_t start;
	uint64_t vaddr;
	uint64_t end;

	if (parse_iomem_single("Kernel code\n", &start, NULL)) {
		fprintf(stderr, "Cannot determine kernel load addr\n");
		return -1;
	}

	vaddr = get_kernel_stext_sym();
	if (!vaddr)
		return -1;

	/*
	 * Need to back up a bit to map swapper_pg_dir
	 */
	start -= 0x2000;
	vaddr -= 0x2000;

	elf_info->kern_paddr_start = start;
	elf_info->kern_vaddr_start = vaddr;

	if (parse_iomem_single("Kernel bss\n", NULL, &end)) {
		fprintf(stderr, "Cannot determine kernel size\n");
		return -1;
	}
	elf_info->kern_size = end - start;

	dbgprintf("kernel_vaddr = 0x%llx paddr 0x%llx\n",
		  elf_info->kern_vaddr_start, elf_info->kern_paddr_start);
	dbgprintf("kernel size = 0x%lx\n", elf_info->kern_size);

	return 0;
}

/* Converts unsigned long to ascii string. */
static void ultoa(unsigned long i, char *str)
{
	int j = 0, k;
	char tmp;

	do {
		str[j++] = i % 10 + '0';
	} while ((i /=10) > 0);
	str[j] = '\0';

	/* Reverse the string. */
	for (j = 0, k = strlen(str) - 1; j < k; j++, k--) {
		tmp = str[k];
		str[k] = str[j];
		str[j] = tmp;
	}
}

/* Adds the elfcorehdr= command line parameter to command line. */
static int cmdline_add_elfcorehdr(char *cmdline, unsigned long addr)
{
	int cmdlen, len, align = 1024;
	char str[30], *ptr;

	/* Passing in elfcorehdr=xxxK format. Saves space required in cmdline.
	 * Ensure 1K alignment*/
	if (addr%align)
		return -1;
	addr = addr/align;
	ptr = str;
	strcpy(str, " elfcorehdr=");
	ptr += strlen(str);
	ultoa(addr, ptr);
	strcat(str, "K");
	len = strlen(str);
	cmdlen = strlen(cmdline) + len;
	if (cmdlen > (COMMAND_LINE_SIZE - 1))
		die("Command line overflow\n");
	strcat(cmdline, str);

	dbgprintf("Command line after adding elfcorehdr\n");
	dbgprintf("%s\n", cmdline);

	return 0;
}

/* determine max_phys_bits which is used to calculate PAGE_OFFSET */
static unsigned long get_max_phys_bits(void)
{
	char line[MAX_LINE];
	unsigned long max_phys_bits = 0;
	char type[MAX_LINE];
	char value1[MAX_LINE];
	char value2[MAX_LINE];
	FILE *fp;

	fp = fopen("/proc/cpuinfo", "r");
	if (!fp) {
		fprintf(stderr, "Cannot open /proc/cpuinfo\n");
		return 0;
	}

	while (fgets(line, sizeof(line), fp) != NULL) {
		sscanf(line, "%s : %s %s", type, value1, value2);
		if (strcmp(type, "cpu"))
			continue;
		if (strcmp(value1, "UltraSparc") == 0) {
			if ((strcmp(value2, "T1") == 0) ||
			    (strcmp(value2, "T2") == 0))
				max_phys_bits = 39;
			else if (strcmp(value2, "T3") == 0)
				max_phys_bits = 43;
			else if ((strcmp(value2, "T4") == 0) ||
				 (strcmp(value2, "T5") == 0))
				max_phys_bits = 47;
			else
				fprintf(stderr, "Unknown cpu %s %s\n", value1, value2);
		} else if (strcmp(value1, "SPARC64-X") == 0)
			max_phys_bits = 47;
		else if (strcmp(value1, "SPARC-M7") == 0)
			max_phys_bits = 49;
		else
			fprintf(stderr, "Unknown cpu %s %s\n", value1, value2);
		break;
	}
	fclose(fp);
	return max_phys_bits;
}

/* Loads additional segments in case of a panic kernel is being loaded.
 * One segment for storing elf headers for crash memory image.
 */
int load_crashdump_segments(struct kexec_info *info, char* mod_cmdline,
				unsigned long max_addr, unsigned long min_base)
{
	void *tmp;
	unsigned long bufsz, memsz, elfcorehdr;
	int nr_ranges = 0, align = 1024;
	struct memory_range *mem_range;
	struct crash_elf_info elf_info;
	unsigned long max_phys_bits = get_max_phys_bits();
/*
	int nr_memmap = 0, i;
	struct memory_range *memmap_p;
	unsigned long sz;
	uint64_t start;
 */

	memset(&elf_info, 0x0, sizeof(elf_info));

	/* Constant parts of the elf_info */
	memset(&elf_info, 0, sizeof(elf_info));
	elf_info.data = ELFDATA2MSB;
	elf_info.machine = EM_SPARCV9;
	elf_info.class = ELFCLASS64;

	if (!max_phys_bits)
		die("Cannot determine PAGE_OFFSET\n");

	elf_info.page_offset = -(1UL << max_phys_bits);

	if (get_crash_memory_ranges(&mem_range, &nr_ranges) || !nr_ranges)
		die("Failed to get crash ranges\n");

	if (get_kernel_region(&elf_info))
		return -1;
/*
	if (parse_iomem_single("Kernel code\n", &start, NULL) == 0) {
		elf_info.kern_paddr_start = start;
		dbgprintf("kernel load physical addr start = 0x%lx\n", start);
	}

	if (get_kernel_vaddr_and_size(info, &elf_info))
		return -1;
 */

	/* Memory regions which panic kernel can safely use to boot into */
/*
	sz = (sizeof(struct memory_range) * CRASH_MAX_MEMMAP_NR);
	memmap_p = xmalloc(sz);
	memset(memmap_p, 0, sz);
	sz = crash_reserved_mem.end - crash_reserved_mem.start + 1;
	if (add_memmap(memmap_p, &nr_memmap, crash_reserved_mem.start, sz, RANGE_RAM) < 0)
		return -1;
 */

	/* Create elf header segment and store crash image data. */
	if (crash_create_elf64_headers(info, &elf_info, mem_range, nr_ranges,
				       &tmp, &bufsz, ELF_CORE_HEADER_ALIGN) < 0)
		return -1;
	/* the size of the elf headers allocated is returned in 'bufsz' */

	memsz = bufsz;
	min_base = crash_reserved_mem.start;
	max_addr = crash_reserved_mem.end;
	elfcorehdr = add_buffer(info, tmp, bufsz, memsz, align, min_base,
							max_addr, -1);
	dbgprintf("Created elf header segment at 0x%lx\n", elfcorehdr);
/*
	if (delete_memmap(memmap_p, &nr_memmap, elfcorehdr, memsz) < 0)
		return -1;
	if (arch_options.pass_memmap_cmdline)
		cmdline_add_memmap(mod_cmdline, memmap_p);
 */
	cmdline_add_elfcorehdr(mod_cmdline, elfcorehdr);

	/* Store 2nd kernel boot memory ranges for later reference in
	 * x86-setup-linux.c: setup_linux_system_parameters() */
/*
	info->crash_range = memmap_p;
	info->nr_crash_ranges = nr_memmap;
 */

	return 0;
}
