#define _GNU_SOURCE
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <limits.h>
#include <elf.h>
#include <boot/elf_boot.h>
#include <ip_checksum.h>
#include "../../kexec.h"
#include "../../kexec-syscall.h"
#include "../../kexec-elf.h"
#include "kexec-sparc64.h"
#include "crashdump-sparc64.h"
#include <arch/options.h>

/*
 * Copied from silo source.
 */
struct HdrS_struct {
	char magic[4];
	unsigned int linux_ver;
	unsigned short ver;
	unsigned short root_flags;
	unsigned short root_dev;
	unsigned short ram_flags;
	unsigned int ramdisk_image;
	unsigned int ramdisk_size;
	/* 0x201 */
	unsigned long long reboot_cmd_ptr;
	/* 0x202 */
	unsigned long long bootstr_info_ptr;
	/* 0x301 */
	unsigned long long ramdisk_image64;
	/* 0x302 */
	unsigned long long sparc_crash_base;
	unsigned long long sparc_crash_size;
	unsigned int sparc_kexec;
};

struct HdrS_struct *find_linux_HdrS(const char *base, long len)
{
    /* Ugly magic to find HdrS, we dig into first jmp gokernel */
    const char *p = base + ((*(unsigned short *) (base+2)) << 2) - 512;
    const char *q;

    q = base+8;
    if (*q == 'H' && q[1] == 'd' && q[2] == 'r' && q[3] == 'S')
	return (struct HdrS_struct *)q;
    if (p >= base + len || p <= base)
	return 0;
    for (q = p + 512; p < q; p += 4) {
	if (*p == 'H' && p[1] == 'd' && p[2] == 'r' && p[3] == 'S')
	    return (struct HdrS_struct *)p;
    }
    return 0;
}

/*
 * A HUGE word of caution on the initrd base address:
 *  silo, and grub2 followed the example, compute this
 *  physical base with (phys-of-initrd minus -phy-base-of-available-mem) plus
 *    KERNBASE.
 */

static void adjust_initrd_location(struct kexec_info *info,
	struct mem_ehdr *ehdr, unsigned long *ram_location)
{
	unsigned long new_ram_location;

	new_ram_location = *ram_location - info->memory_range[0].start +
		KERNBASE;
	fprintf(stderr, "%s: ram_location = 0x%lx new_ram_location = 0x%lx\n",
		__FUNCTION__, *ram_location, new_ram_location);
	*ram_location = new_ram_location;
}

/* This is another hardcoded assumption about the kernel.
 * The boot args layout is known by silo, grub2 and the kernel.
 * It is now known by kexec too. We might need a better way to
 * do this but without compromising the boot loaders.
 */
struct bootstr_info {
	int bootstr_len;
	int bootstr_valid;
	char bootstr_buf[0];
};

int elf_sparc64_load(int argc, char **argv, const char *buf, off_t len,
	struct kexec_info *info)
{
	unsigned long base = 0ul;
	off_t initrd_size = 0ul;
	const char *command_line = NULL;
	char *modified_cmdline = NULL;
	off_t command_line_len = (off_t) 0;
	const char *ramdisk = NULL;
	char *ramdisk_buf = NULL;
	struct mem_ehdr ehdr;
	unsigned long max_addr;
	int opt, result;
	int crash_ranges = 0;
	uint64_t start;
	uint64_t end = ULONG_MAX;
	struct HdrS_struct *HdrS;
	int rc;

	/* See options.h -- add any more there, too. */
	static const struct option options[] = {
		KEXEC_ARCH_OPTIONS
		{ "command-line",	1, 0, OPT_APPEND },
		{ "append",		1, 0, OPT_APPEND },
		{ "initrd",		1, 0, OPT_RAMDISK },
		{ 0, 			0, 0, 0 },
	};
	static const char short_options[] = KEXEC_ARCH_OPT_STR "";

	while((opt = getopt_long(argc, argv, short_options, options, 0)) != -1){
		switch(opt) {
		default:
			if (opt < OPT_ARCH_MAX) {
				break;
			}
		case '?':
			usage();
			return -1;
		case OPT_APPEND:
			command_line = optarg;
			break;
		case OPT_RAMDISK:
			ramdisk = optarg;
			break;
		}
	}

	result = build_elf_exec_info(buf, len, &ehdr, 0);
	if (result < 0)
		die("ELF exec parse error\n");

	max_addr = elf_max_addr(&ehdr);

	if (command_line) {
		command_line_len = strlen(command_line) + 1;
		if (command_line_len > COMMAND_LINE_SIZE)
			command_line_len = COMMAND_LINE_SIZE;
	}

	if (info->kexec_flags & KEXEC_ON_CRASH) {
		modified_cmdline = xmalloc(COMMAND_LINE_SIZE);
		if (!modified_cmdline)
			return -1;

		if (command_line) {
			(void) strncpy(modified_cmdline, command_line,
				       COMMAND_LINE_SIZE);
			modified_cmdline[COMMAND_LINE_SIZE - 1] = '\0';
		}

		rc = load_crashdump_segments(info, modified_cmdline, max_addr,
					     0);
		if (rc < 0)
			return -1;

		command_line = modified_cmdline;
		command_line_len = strlen(command_line) + 1;

		/* Move the PT_LOAD p_paddr-s. */
		kexec_crash_move_pt_load(&ehdr);

		/*
		 * We put the dump capture kernel at the start of crashkernel
		 * reserved memory.
		 */
		if (parse_iomem_single("Crash kernel\n", &start, &end)) {
			/*
			 * No crash kernel memory reserved. We cannot do more
			 * but just bail out.
			 */
			return -1;
		}
		base = start;
	} else
		kexec_move_pt_load(&ehdr);

	if (base == ULONG_MAX)
		return -1;

	result = elf_exec_load(&ehdr, info);
	if (result < 0)
		die("ELF exec load\n");

	/* Make sure kernel is segment 0 */
	sort_segments(info);

	HdrS = find_linux_HdrS(info->segment[0].buf + 0x4000,
			       info->segment[0].memsz);
	if (!HdrS)
		die("kernel HdrS structure not found\n");
	if (HdrS->ver < 0x302)
		die("invalid kernel HdrS version 0x%x\n", HdrS->ver);

	if (ramdisk)
		ramdisk_buf = slurp_file(ramdisk, &initrd_size);

	if (ramdisk_buf) {
		unsigned long ram_location;

		ram_location = add_buffer(info, ramdisk_buf, initrd_size,
			initrd_size, HPAGE_SIZE, base, end, -1);
		adjust_initrd_location(info, &ehdr, &ram_location);

		HdrS->ramdisk_image64 = ram_location;
		HdrS->ramdisk_size = initrd_size;
	}

	if (crash_reserved_mem.start) {
		HdrS->sparc_crash_base = crash_reserved_mem.start;
		HdrS->sparc_crash_size = crash_reserved_mem.end -
					 crash_reserved_mem.start + 1;
	}

	if (command_line) {
		struct bootstr_info *p_bootstr_info;

		p_bootstr_info = (struct bootstr_info *)(info->segment[0].buf +
				  HdrS->bootstr_info_ptr - KERNBASE);
		strncpy(p_bootstr_info->bootstr_buf, command_line,
			p_bootstr_info->bootstr_len);
		p_bootstr_info->bootstr_valid = 1;
	}

	/* This assumes the kernel is mapped at KERNBASE */
	info->entry = (void *) info->segment[0].mem + ehdr.e_entry - KERNBASE;
	HdrS->sparc_kexec = 1;
	free_elf_info(&ehdr);
	return 0;
}
